﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration.GrammarEnumeration;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {
  [Item("Grammar Enumeration Symbolic Regression", "Iterates all possible model structures for a fixed grammar.")]
  [StorableClass]
  [Creatable(CreatableAttribute.Categories.DataAnalysisRegression, Priority = 250)]
  public class GrammarEnumerationAlgorithm : FixedDataAnalysisAlgorithm<IRegressionProblem> {
    public override IDeepCloneable Clone(Cloner cloner) {
      return new GrammarEnumerationAlgorithm(this, cloner);
    }


    public GrammarEnumerationAlgorithm() {
      Problem = new RegressionProblem();

    }


    private GrammarEnumerationAlgorithm(GrammarEnumerationAlgorithm original, Cloner cloner) : base(original, cloner) { }





    protected override void Run(CancellationToken cancellationToken) {

      IntValue generatedSolutions = new IntValue();
      Results.Add(new Result("Generated Solutions", generatedSolutions));

      DoubleValue averageTreeLength = new DoubleValue();
      Results.Add(new Result("Average Tree Length of Solutions", averageTreeLength));

      int maxStringLength = 4;


      List<SymbolString> results = new List<SymbolString>();

      Grammar grammar = new Grammar(new[] { "a", "b" });

      Stack<SymbolString> remainingTrees = new Stack<SymbolString>();
      remainingTrees.Push(new SymbolString(new[] { grammar.StartSymbol }));

      while (remainingTrees.Any()) {
        SymbolString currSymbolString = remainingTrees.Pop();

        if (currSymbolString.IsSentence()) {
          results.Add(currSymbolString);

          generatedSolutions.Value++;
          averageTreeLength.Value = results.Select(r => r.Count).Average();

        } else {
          // expand next nonterminal symbols
          int nonterminalSymbolIndex = currSymbolString.FindIndex(s => s is NonterminalSymbol);

          NonterminalSymbol expandedSymbol = currSymbolString[nonterminalSymbolIndex] as NonterminalSymbol;

          foreach (Production productionAlternative in expandedSymbol.Alternatives) {
            SymbolString newSentence = new SymbolString(currSymbolString);
            newSentence.RemoveAt(nonterminalSymbolIndex);
            newSentence.InsertRange(nonterminalSymbolIndex, productionAlternative);

            if (newSentence.Count <= maxStringLength) {
              remainingTrees.Push(newSentence);
            }
          }
        }
      }


      StringArray sentences = new StringArray(results.Select(r => r.ToString()).ToArray());
      Results.Add(new Result("All sentences", sentences));



      /*
      addSy = grammar.AllowedSymbols.OfType<Addition>().First();
      mulSy = grammar.AllowedSymbols.OfType<Multiplication>().First();
      logSy = grammar.AllowedSymbols.OfType<Logarithm>().First();
      expSy = grammar.AllowedSymbols.OfType<Exponential>().First();
      divSy = grammar.AllowedSymbols.OfType<Division>().First();

      progRootSy = grammar.AllowedSymbols.OfType<ProgramRootSymbol>().First();
      startSy = grammar.AllowedSymbols.OfType<StartSymbol>().First();



      var rootNode = progRootSy.CreateTreeNode();

      var startNode = startSy.CreateTreeNode();
      rootNode.AddSubtree(startNode);

      startNode.AddSubtree(const0);



      SymbolicExpressionTree tree = new SymbolicExpressionTree(rootNode);

      SymbolicRegressionModel model = new SymbolicRegressionModel(
        Problem.ProblemData.TargetVariable,
        tree,
        new SymbolicDataAnalysisExpressionTreeLinearInterpreter());

      IRegressionSolution solution = model.CreateRegressionSolution(Problem.ProblemData);
      Results.Add(new Result("Best solution (training)", solution));
      Results.Add(new Result("Best solution quality (training)", new DoubleValue(solution.TrainingRSquared).AsReadOnly()));

      Results.Add(new Result("Best solution (test)", solution));
      Results.Add(new Result("Best solution quality (test)", new DoubleValue(solution.TestRSquared).AsReadOnly()));
      */
    }
  }
}